// Includes -----------------------------------------------------------------
#include <SPI.h>
#include <Wire.h>
#include <LiquidCrystal_I2C.h>

// Ethernet 
#include <Ethernet2.h>

// PubSubClient
#include <PubSubClient.h>

// Defines -----------------------------------------------------------------
// Display
#define BLANK  (char)32 // ASCII Leerzeichen
#define EMPTY  (char)48 // ASCII 0
#define ENDE   (char)0  // Stringende
#define PUNKT  (char)46 // ASCII .
#define DPUNKT (char)58 // ASCII :
#define ZWEI   (char)50 // ASCII 2

// Fehlermeldungen
#define ERR_OK     (char)0
#define ERR_NODHCP (char)1
#define ERR_ETH    (char)3
#define ERR_MQTT   (char)4
#define ERR_MQTTOK (char)5

// Variablen -----------------------------------------------------------------
LiquidCrystal_I2C lcd(0x3F, 2, 1, 0, 4, 5, 6, 7, 3, POSITIVE);

// MQTT
char tmpstr[3] = {EMPTY,EMPTY,ENDE};
char humstr[3] = {EMPTY,EMPTY,ENDE};
char datum[11] = {EMPTY,EMPTY,PUNKT,EMPTY,EMPTY,PUNKT,ZWEI,EMPTY,EMPTY,EMPTY,ENDE};
char zeit[6]   = {EMPTY,EMPTY,DPUNKT,EMPTY,EMPTY,ENDE};
const char top1[] = "Wohnung/Raum1/Temperatur/Messwert";
const char top2[] = "Wohnung/Raum1/Temperatur/Messzeit";
const char top3[] = "Wohnung/Raum1/Luftfeuchte/Messwert";
const char top4[] = "Wohnung/Raum1/Luftfeuchte/Messzeit";
bool neuerwert = false;

// Ethernet
byte mac[] = { 0x12,0x34,0x56,0x78,0x9A,0xBE };
EthernetClient client;

// PubSubClient
IPAddress    pscsrv(192, 168, 188, 25);
unsigned int pscport = 1883;
PubSubClient pscclient(client);

// Funktionen ----------------------------------------------------------------
// MQTT
void MQTTReconnect() 
{
  // Loop until we're reconnected
  while (!pscclient.connected()) 
  {
    Serial.print("MQTT Verbindungsaufbau... ");
    if (pscclient.connect("JGE-Sub")) 
    {
      Serial.println("MQTT verbunden");
      pscclient.subscribe(top1);
      pscclient.subscribe(top2);
      pscclient.subscribe(top3);
      pscclient.subscribe(top4);
      LCD_ShowFehler(ERR_MQTTOK);
    } 
    else 
    {
      Serial.print(F("MQTT Fehler, Status = "));
      Serial.print(pscclient.state());
      Serial.println(" Naechster Versuch in 10 Sekunden");
      LCD_ShowFehler(ERR_MQTT);
      delay(5000);
    }
  }
}

void MQTTCallback(char* topic, byte* payload, unsigned int length) 
{
  int i = 0;
  
  Serial.print("Nachricht angekommen [");
  Serial.print(topic);
  Serial.print("] ");
  for (i = 0; i < length; i++) 
  {
    Serial.print((char)payload[i]);
  }
  Serial.println();
  if (strcmp(topic,top1) == 0)
  {
    tmpstr[0] = payload[0];
    tmpstr[1] = payload[1];
  }
  if (strcmp(topic,top2) == 0)
  {
    datum[0] = payload[0];
    datum[1] = payload[1];
    datum[3] = payload[3];
    datum[4] = payload[4];
    datum[8] = payload[8];
    datum[9] = payload[9];

    zeit[0] = payload[11];
    zeit[1] = payload[12];
    zeit[3] = payload[14];
    zeit[4] = payload[15];
  }
  if (strcmp(topic,top3) == 0)
  {
    humstr[0] = payload[0];
    humstr[1] = payload[1];
  }  
  if (strcmp(topic,top4) == 0)
  {
    datum[0] = payload[0];
    datum[1] = payload[1];
    datum[3] = payload[3];
    datum[4] = payload[4];
    datum[8] = payload[8];
    datum[9] = payload[9];

    zeit[0] = payload[11];
    zeit[1] = payload[12];
    zeit[3] = payload[14];
    zeit[4] = payload[15];
  }
  neuerwert = true;
}

// LCD
void LCD_ShowValues(void)
{
  char grad[2] = {(char)223,(char)0}; // Gradzeichen nach Datenblatt LCD 44780

  lcd.clear();
  lcd.setCursor(0,0);
  lcd.print(F("Temperatur  = "));
  lcd.print(tmpstr);
  lcd.print(grad);
  lcd.print(F("C"));
  lcd.setCursor(0,1);
  lcd.print(F("Luftfeuchte = "));
  lcd.print(humstr);
  lcd.print(F("%"));
  lcd.setCursor(0,2);
  lcd.print(F("Datum   : "));
  lcd.print(datum);
  lcd.setCursor(0,3);
  lcd.print(F("Uhrzeit : "));
  lcd.print(zeit);
}

void LCD_ShowFehler(char fno)
{
  lcd.clear();
  lcd.setCursor(0,0);
  lcd.print(F("Fehler:"));
  switch(fno)
  {
    case ERR_OK :
      lcd.setCursor(0,0);
      lcd.print(F("Ethernet Verbunden!"));
      lcd.setCursor(0,1);
      lcd.print(F("Alles Gut"));
      lcd.setCursor(0,2);
      lcd.print(F(":-):-):-)"));
      break;
    case ERR_MQTTOK :
      lcd.setCursor(0,0);
      lcd.print(F("MQTT Verbunden!"));
      lcd.setCursor(0,1);
      lcd.print(F("Warte auf Messwerte"));
      lcd.setCursor(0,2);
      lcd.print(F("....."));
      break;
    case ERR_NODHCP :
      lcd.setCursor(0,1);
      lcd.print(F("kein DHCP!"));
      lcd.setCursor(0,2);
      lcd.print(F("Netzwerk ?"));
      break;
    case ERR_ETH :
      lcd.setCursor(0,1);
      lcd.print(F("Ethernet!"));
      lcd.setCursor(0,2);
      lcd.print(F("Controller Init"));
      break;
    case ERR_MQTT :
      lcd.setCursor(0,1);
      lcd.print(F("MQTT Server!"));
      lcd.setCursor(0,2);
      lcd.print(F("keine verbindung ..."));
      lcd.setCursor(0,3);
      lcd.print(F("Wiederhole in 10 Sek"));
      break;
    default:
      lcd.setCursor(0,1);
      lcd.print(F("unbekannt!"));
      lcd.setCursor(0,2);
      lcd.print(F("??????????"));
      break;
  }
  delay(5000);
}

void setup() 
{
  Serial.begin(9600);
  Serial.println(F("[MQTT-Subscriber-LCD]"));

  // OLED
  lcd.begin(20,4);
  lcd.clear();
  lcd.setCursor(0,0);
  lcd.print(F("+------------------+"));
  lcd.setCursor(0,1);
  lcd.print(F("|      SEaP:       |"));
  lcd.setCursor(0,2);
  lcd.print(F("|                  |"));
  lcd.setCursor(0,3);
  lcd.print(F("+------------------+"));
  delay(2000);

  // MQTT
  pscclient.setServer(pscsrv, pscport);
  pscclient.setCallback(MQTTCallback);

  // Ethernet
  Serial.println(F("Starte Ethernet Init"));
  if (Ethernet.begin(mac) == 0)
  {
    Serial.println(F("Ethernet initialisierung fehler"));
    LCD_ShowFehler(ERR_ETH);
    delay(10000);
  }
  else
  {
    LCD_ShowFehler(ERR_OK);
    Serial.println(Ethernet.localIP());
    LCD_ShowValues(); // Maske mit Initwerten anzeigen
  }
}

void loop () 
{
  // MQTT
  if (!pscclient.connected()) 
  {
    MQTTReconnect();
  }
  pscclient.loop();

  // LCD
  if (true == neuerwert)
  {
    LCD_ShowValues();
    neuerwert = false;
  }
}

